"""Perfect predictors."""
from dataclasses import dataclass
from typing import ClassVar

from ranzen import implements

from ethicml.algorithms.inprocess.in_algorithm import InAlgorithmNoParams
from ethicml.utility import DataTuple, Prediction, TestTuple

from ..postprocess.dp_flip import DPFlip

__all__ = ["Oracle", "DPOracle"]


@dataclass
class Oracle(InAlgorithmNoParams):
    """A perfect predictor.

    Can only be used if test is a DataTuple, rather than the usual TestTuple.
    This model isn't intended for general use,
    but can be useful if you want to either do a sanity check, or report potential values.
    """

    is_fairness_algo: ClassVar[bool] = False

    @implements(InAlgorithmNoParams)
    def get_name(self) -> str:
        return "Oracle"

    @implements(InAlgorithmNoParams)
    def fit(self, train: DataTuple, seed: int = 888) -> "Oracle":
        return self

    @implements(InAlgorithmNoParams)
    def predict(self, test: TestTuple) -> Prediction:
        assert isinstance(test, DataTuple), "test must be a DataTuple."
        return Prediction(hard=test.y.copy())

    @implements(InAlgorithmNoParams)
    def run(self, train: DataTuple, test: TestTuple, seed: int = 888) -> Prediction:
        assert isinstance(test, DataTuple), "test must be a DataTuple."
        return Prediction(hard=test.y.copy())


@dataclass
class DPOracle(InAlgorithmNoParams):
    """A perfect Demographic Parity Predictor.

    Can only be used if test is a DataTuple, rather than the usual TestTuple.
    This model isn't intended for general use,
    but can be useful if you want to either do a sanity check, or report potential values.
    """

    @implements(InAlgorithmNoParams)
    def get_name(self) -> str:
        return "DemPar. Oracle"

    @implements(InAlgorithmNoParams)
    def fit(self, train: DataTuple, seed: int = 888) -> "DPOracle":
        self.seed = seed
        return self

    @implements(InAlgorithmNoParams)
    def predict(self, test: TestTuple) -> Prediction:
        assert isinstance(test, DataTuple), "test must be a DataTuple."
        flipper = DPFlip()
        test_preds = Prediction(test.y.copy())
        return flipper.run(test_preds, test, test_preds, test, seed=self.seed)

    @implements(InAlgorithmNoParams)
    def run(self, train: DataTuple, test: TestTuple, seed: int = 888) -> Prediction:
        assert isinstance(test, DataTuple), "test must be a DataTuple."
        flipper = DPFlip()
        test_preds = Prediction(test.y.copy())
        return flipper.run(test_preds, test, test_preds, test, seed=seed)



"""Installable model.

This is a kind of complicated model, but it's incredibly useful.
Say you find a paper from a few years ago with code. It's not unreasonable that there might
be dependency clashes, python clashes, clashes galore. This approach downloads a model, runs it
in its own venv and makes everyone happy.
"""
import os
import shutil
import subprocess
import sys
from abc import ABC
from pathlib import Path
from typing import Optional

import git
from ranzen.decorators import implements

from ethicml.algorithms.inprocess.in_algorithm import InAlgorithm

from ..algorithm_base import SubprocessAlgorithmMixin

__all__ = ["InstalledModel"]


class InstalledModel(SubprocessAlgorithmMixin, InAlgorithm, ABC):
    """The model that does the magic.

    Download code from given URL and create Pip environment with Pipfile found in the code.

    :param name: Name of the model.
    :param dir_name: Where to download the code to (can be chosen freely).
    :param top_dir: Top directory of the repository where the Pipfile can be found (this is usually
        simply the last part of the repository URL).
    :param url: URL of the repository. (Default: None)
    :param executable: Path to a Python executable. (Default: None.
    :param use_poetry: If True, will try to use poetry instead of pipenv. (Default: False)
    """

    def __init__(
        self,
        name: str,
        dir_name: str,
        top_dir: str,
        url: Optional[str] = None,
        executable: Optional[str] = None,
        use_poetry: bool = False,
    ):
        # QUESTION: do we really need `store_dir`? we could also just clone the code into "."
        self._store_dir: Path = Path(".") / dir_name  # directory where code and venv are stored
        self._top_dir: str = top_dir

        if url is not None:
            # download code
            self._clone_directory(url)

        if executable is None:
            # create virtual environment
            del use_poetry  # see https://github.com/python-poetry/poetry/issues/4055
            # self._create_venv(use_poetry=use_poetry)
            self._create_venv(use_poetry=False)
            self.__executable = str(self._code_path.resolve() / ".venv" / "bin" / "python")
        else:
            self.__executable = executable
        self.__name = name

    @implements(InAlgorithm)
    def get_name(self) -> str:
        return self.__name

    @property
    def _code_path(self) -> Path:
        """Path to where the code of the model is located."""
        return self._store_dir / self._top_dir

    @property
    def executable(self) -> str:
        """Python executable from the virtualenv associated with the model."""
        return self.__executable

    def _clone_directory(self, url: str) -> None:
        """Clones the repo from `url` into `self._store_dir`.

        :param url: URL of the repository.
        """
        if not self._store_dir.exists():
            self._store_dir.mkdir()
            git.Git(self._store_dir).clone(url)

    def _create_venv(self, use_poetry: bool) -> None:
        """Create a venv based on the Pipfile in the repository.

        :param use_poetry: Whether to use poetry instead of pipenv.
        """
        venv_directory = self._code_path / ".venv"
        if not venv_directory.exists():
            if use_poetry and shutil.which("poetry") is not None:  # use poetry instead of pipenv
                environ = os.environ.copy()
                environ["POETRY_VIRTUALENVS_CREATE"] = "true"
                environ["POETRY_VIRTUALENVS_IN_PROJECT"] = "true"
                subprocess.run(
                    ["poetry", "install", "--no-root"], env=environ, check=True, cwd=self._code_path
                )
                return

            environ = os.environ.copy()
            environ["PIPENV_IGNORE_VIRTUALENVS"] = "1"
            environ["PIPENV_VENV_IN_PROJECT"] = "true"
            environ["PIPENV_YES"] = "true"
            environ["PIPENV_PIPFILE"] = str(self._code_path / "Pipfile")

            subprocess.run([sys.executable, "-m", "pipenv", "install"], env=environ, check=True)

    def remove(self) -> None:
        """Remove the directory that we created in :meth:`_clone_directory()`."""
        try:
            shutil.rmtree(self._store_dir)
        except OSError as excep:
            print(f"Error: {excep.filename} - {excep.strerror}.")



"""Abstract Base Class of all metrics in the framework."""

from abc import ABC, abstractmethod
from typing import ClassVar
from typing_extensions import final

from ranzen import implements

from ethicml.utility import EvalTuple, Prediction

__all__ = ["MetricStaticName", "Metric"]


class Metric(ABC):
    """Base class for all metrics."""

    apply_per_sensitive: ClassVar[bool] = True
    """Whether the metric can be applied per sensitive attribute."""

    @abstractmethod
    def score(self, prediction: Prediction, actual: EvalTuple) -> float:
        """Compute score.

        :param prediction: predicted labels
        :param actual: EvalTuple with the actual labels and the sensitive attributes
        :returns: the score as a single number
        """

    @abstractmethod
    def get_name(self) -> str:
        """Name of the metric."""

    @property
    @final
    def name(self) -> str:
        """Name of the metric."""
        return self.get_name()


class MetricStaticName(Metric, ABC):
    """Metric base class for metrics whose name does not depend on instance variables."""

    _name: ClassVar[str] = "<please overwrite me>"

    @implements(Metric)
    def get_name(self) -> str:
        return self._name



"""Implementation of logistic regression (actually just a wrapper around sklearn)."""
from __future__ import annotations

import contextlib
import json
import os
import random
import sys
from pathlib import Path
from typing import TYPE_CHECKING, Generator

import numpy as np
import pandas as pd
from joblib import dump, load
from sklearn.ensemble import GradientBoostingClassifier
from sklearn.linear_model import LogisticRegression

from ethicml.algorithms.inprocess.svm import select_svm
from ethicml.utility import (
    ClassifierType,
    DataTuple,
    FairnessType,
    KernelType,
    Prediction,
    SubgroupTuple,
    TestTuple,
)

if TYPE_CHECKING:
    from fairlearn.reductions import ExponentiatedGradient

    from ethicml.algorithms.inprocess.agarwal_reductions import AgarwalArgs
    from ethicml.algorithms.inprocess.in_subprocess import InAlgoArgs


def fit(train: DataTuple, args: AgarwalArgs, seed: int = 888) -> ExponentiatedGradient:
    """Fit a model.

    :param train:
    :param args:
    """
    try:
        from fairlearn.reductions import (
            DemographicParity,
            EqualizedOdds,
            ExponentiatedGradient,
            UtilityParity,
        )
    except ImportError as e:
        raise RuntimeError(
            f"In order to use Agarwal, install fairlearn==0.7.0. "
            f"Consider installing EthicML with the extras 'all' specified."
        ) from e

    fairness_class: UtilityParity
    fairness_type = FairnessType[args["fairness"]]
    classifier_type = ClassifierType[args["classifier"]]
    kernel_type = None if args["kernel"] == "" else KernelType[args["kernel"]]

    if fairness_type is FairnessType.dp:
        fairness_class = DemographicParity(difference_bound=args["eps"])
    else:
        fairness_class = EqualizedOdds(difference_bound=args["eps"])

    if classifier_type is ClassifierType.svm:
        assert kernel_type is not None
        model = select_svm(C=args["C"], kernel=kernel_type, seed=seed)
    elif classifier_type is ClassifierType.lr:
        random_state = np.random.RandomState(seed=seed)
        model = LogisticRegression(
            solver="liblinear", random_state=random_state, max_iter=5000, C=args["C"]
        )
    elif classifier_type is ClassifierType.gbt:
        random_state = np.random.RandomState(seed=seed)
        model = GradientBoostingClassifier(random_state=random_state)

    data_x = train.x
    data_y = train.y
    data_a = train.s

    exponentiated_gradient = ExponentiatedGradient(
        model, constraints=fairness_class, eps=args["eps"], max_iter=args["iters"]
    )
    exponentiated_gradient.fit(data_x, data_y, sensitive_features=data_a)

    min_class_label = train.y.min()
    exponentiated_gradient.min_class_label = min_class_label

    return exponentiated_gradient


def predict(exponentiated_gradient: ExponentiatedGradient, test: TestTuple) -> pd.DataFrame:
    """Compute predictions on the given test data.

    :param exponentiated_gradient:
    :param test:
    """
    randomized_predictions = exponentiated_gradient.predict(test.x)
    preds = pd.DataFrame(randomized_predictions, columns=["preds"])

    if preds["preds"].min() != preds["preds"].max():
        preds = preds.replace(preds["preds"].min(), exponentiated_gradient.min_class_label)
    return preds


def train_and_predict(
    train: DataTuple, test: TestTuple, args: AgarwalArgs, seed: int
) -> pd.DataFrame:
    """Train a logistic regression model and compute predictions on the given test data.

    :param train:
    :param test:
    :param args:
    """
    exponentiated_gradient = fit(train, args, seed)
    return predict(exponentiated_gradient, test)


@contextlib.contextmanager
def working_dir(root: Path) -> Generator[None, None, None]:
    """Change the working directory to the given path.

    :param root:
    """
    curdir = os.getcwd()
    os.chdir(root.expanduser().resolve().parent)
    try:
        yield
    finally:
        os.chdir(curdir)


def main() -> None:
    """Run the Agarwal model as a standalone program."""
    in_algo_args: InAlgoArgs = json.loads(sys.argv[1])
    flags: AgarwalArgs = json.loads(sys.argv[2])
    try:
        import cloudpickle

        # Need to install cloudpickle for now. See https://github.com/fairlearn/fairlearn/issues/569
    except ImportError as e:
        raise RuntimeError("In order to use Agarwal, install fairlearn and cloudpickle.") from e

    if in_algo_args["mode"] == "run":
        random.seed(in_algo_args["seed"])
        np.random.seed(in_algo_args["seed"])
        train, test = DataTuple.from_npz(Path(in_algo_args["train"])), SubgroupTuple.from_npz(
            Path(in_algo_args["test"])
        )
        Prediction(
            hard=train_and_predict(train, test, flags, in_algo_args["seed"])["preds"]
        ).to_npz(Path(in_algo_args["predictions"]))
    elif in_algo_args["mode"] == "fit":
        random.seed(in_algo_args["seed"])
        np.random.seed(in_algo_args["seed"])
        data = DataTuple.from_npz(Path(in_algo_args["train"]))
        model = fit(data, flags, in_algo_args["seed"])
        with working_dir(Path(in_algo_args["model"])):
            model.ethicml_random_seed = in_algo_args["seed"]  # need to save the seed as well
            model_file = cloudpickle.dumps(model)
        dump(model_file, Path(in_algo_args["model"]))
    elif in_algo_args["mode"] == "predict":
        testdata = SubgroupTuple.from_npz(Path(in_algo_args["test"]))
        model_file = load(Path(in_algo_args["model"]))
        with working_dir(Path(in_algo_args["model"])):
            model = cloudpickle.loads(model_file)
            seed = model.ethicml_random_seed
        random.seed(seed)
        np.random.seed(seed)
        Prediction(hard=predict(model, testdata)["preds"]).to_npz(Path(in_algo_args["predictions"]))
    else:
        raise RuntimeError(f"Unknown mode: {in_algo_args['mode']}")


if __name__ == "__main__":
    main()



"""Applies sci-kit learn's confusion matrix."""

from dataclasses import dataclass
from typing import List, Optional, Tuple

import numpy as np
from sklearn.metrics import confusion_matrix as conf_mtx

from ethicml.metrics.metric import MetricStaticName
from ethicml.utility.data_structures import EvalTuple, Prediction

__all__ = ["CfmMetric", "LabelOutOfBounds"]


@dataclass  # type: ignore[misc]  # mypy doesn't allow abstract dataclasses because mypy is stupid
class CfmMetric(MetricStaticName):
    """Confusion Matrix based metric."""

    pos_class: int = 1
    """The class to treat as being "positive"."""
    labels: Optional[List[int]] = None
    """List of possible target values. If `None`, then this is inferred from the data when run."""

    def confusion_matrix(
        self, prediction: Prediction, actual: EvalTuple
    ) -> Tuple[int, int, int, int]:
        """Apply sci-kit learn's confusion matrix.

        :param prediction: The predictions.
        :param actual: The actual labels.
        """
        actual_y: np.ndarray = actual.y.to_numpy(dtype=np.int32)  # type: ignore[type-var]
        _labels: np.ndarray = np.unique(actual_y) if self.labels is None else np.array(self.labels)
        if _labels.size == 1:
            _labels = np.array([0, 1], dtype=np.int32)
        conf_matr: np.ndarray = conf_mtx(y_true=actual_y, y_pred=prediction.hard, labels=_labels)

        if self.pos_class not in _labels:
            raise LabelOutOfBounds("Positive class specified must exist in the test set")

        tp_idx: np.int64 = (_labels == self.pos_class).nonzero()[0].item()
        tp_idx = int(tp_idx)

        true_pos = conf_matr[tp_idx, tp_idx]
        false_pos = conf_matr[:, tp_idx].sum() - true_pos
        false_neg = conf_matr[tp_idx, :].sum() - true_pos
        true_neg = conf_matr.sum() - true_pos - false_pos - false_neg
        return true_neg, false_pos, false_neg, true_pos


class LabelOutOfBounds(Exception):
    """Metric Not Applicable per sensitive attribute, apply to whole dataset instead."""



"""Post-processing method by Hardt et al."""
from dataclasses import dataclass

import numpy as np
import pandas as pd
from numpy.random import RandomState
from ranzen import implements
from scipy.optimize import OptimizeResult, linprog  # type: ignore[attr-defined]

from ethicml.metrics.per_sensitive_attribute import metric_per_sensitive_attribute
from ethicml.metrics.tnr import TNR
from ethicml.metrics.tpr import TPR
from ethicml.utility import DataTuple, Prediction, TestTuple

from .post_algorithm import PostAlgorithm

__all__ = ["Hardt"]


@dataclass
class Hardt(PostAlgorithm):
    """Post-processing method by Hardt et al."""

    unfavorable_label: int = 0
    favorable_label: int = 1

    @implements(PostAlgorithm)
    def get_name(self) -> str:
        return "Hardt"

    @implements(PostAlgorithm)
    def fit(self, train_predictions: Prediction, train: DataTuple) -> PostAlgorithm:
        self.model_params = self._fit(train_predictions, train)
        return self

    @implements(PostAlgorithm)
    def predict(self, test_predictions: Prediction, test: TestTuple, seed: int = 888) -> Prediction:
        return self._predict(self.model_params, test_predictions, test, seed)

    @implements(PostAlgorithm)
    def run(
        self,
        train_predictions: Prediction,
        train: DataTuple,
        test_predictions: Prediction,
        test: TestTuple,
        seed: int = 888,
    ) -> Prediction:
        model_params = self._fit(train_predictions, train)
        return self._predict(model_params, test_predictions, test, seed)

    def _fit(self, train_predictions: Prediction, train: DataTuple) -> OptimizeResult:
        # compute basic statistics
        fraction_s0 = (train.s.to_numpy() == 0).mean()
        fraction_s1 = 1 - fraction_s0

        s_col = train.s.name
        tprs = metric_per_sensitive_attribute(train_predictions, train, TPR())
        tpr0 = tprs[f"{s_col}_0"]
        tpr1 = tprs[f"{s_col}_1"]
        fnr0 = 1 - tpr0
        fnr1 = 1 - tpr1
        tnrs = metric_per_sensitive_attribute(train_predictions, train, TNR())
        tnr0 = tnrs[f"{s_col}_0"]
        tnr1 = tnrs[f"{s_col}_1"]
        fpr0 = 1 - tnr0
        fpr1 = 1 - tnr1

        # linear program has 4 decision variables:
        # [P[label_tilde = 1 | label_hat = 1, sensitive_attribute = 0];
        #  P[label_tilde = 1 | label_hat = 0, sensitive_attribute = 0];
        #  P[label_tilde = 1 | label_hat = 1, sensitive_attribute = 1];
        #  P[label_tilde = 1 | label_hat = 0, sensitive_attribute = 1]]
        # Coefficients of the linear objective function to be minimized.
        coeffs = np.array([fpr0 - tpr0, tnr0 - fnr0, fpr1 - tpr1, tnr1 - fnr1])

        # inequalilty_constraint_matrix: 2-D array which, when matrix-multiplied by x, gives the
        # values of the upper-bound inequality constraints at x
        # b_ub: 1-D array of values representing the upper-bound of each
        # inequality constraint (row) in A_ub.
        # Just to keep these between zero and one
        inequalilty_constraint_matrix: np.ndarray = np.array(
            [
                [1.0, 0.0, 0.0, 0.0],
                [-1.0, 0.0, 0.0, 0.0],
                [0.0, 1.0, 0.0, 0.0],
                [0.0, -1.0, 0.0, 0.0],
                [0.0, 0.0, 1.0, 0.0],
                [0.0, 0.0, -1.0, 0.0],
                [0.0, 0.0, 0.0, 1.0],
                [0.0, 0.0, 0.0, -1.0],
            ],
            dtype=np.float64,
        )
        b_ub: np.ndarray = np.array([1.0, 0.0, 1.0, 0.0, 1.0, 0.0, 1.0, 0.0], dtype=np.float64)

        # Create boolean conditioning vectors for protected groups
        mask_s1 = train.s.to_numpy() == 1
        mask_s0 = train.s.to_numpy() == 0

        train_preds_numpy: np.ndarray = train_predictions.hard.to_numpy()

        sconst = np.ravel(train_preds_numpy[mask_s1] == self.favorable_label)
        sflip = np.ravel(train_preds_numpy[mask_s1] == self.unfavorable_label)
        oconst = np.ravel(train_preds_numpy[mask_s0] == self.favorable_label)
        oflip = np.ravel(train_preds_numpy[mask_s0] == self.unfavorable_label)

        y_true = train.y.to_numpy().ravel()

        sm_tn = np.logical_and(sflip, y_true[mask_s1] == self.unfavorable_label)
        sm_fn = np.logical_and(sflip, y_true[mask_s1] == self.favorable_label)
        sm_fp = np.logical_and(sconst, y_true[mask_s1] == self.unfavorable_label)
        sm_tp = np.logical_and(sconst, y_true[mask_s1] == self.favorable_label)

        om_tn = np.logical_and(oflip, y_true[mask_s0] == self.unfavorable_label)
        om_fn = np.logical_and(oflip, y_true[mask_s0] == self.favorable_label)
        om_fp = np.logical_and(oconst, y_true[mask_s0] == self.unfavorable_label)
        om_tp = np.logical_and(oconst, y_true[mask_s0] == self.favorable_label)

        # A_eq - 2-D array which, when matrix-multiplied by x,
        # gives the values of the equality constraints at x.
        # b_eq - 1-D array of values representing the RHS of each equality constraint (row) in A_eq.
        # Used to impose equality of odds constraint
        a_eq = [
            [
                (np.mean(sconst * sm_tp) - np.mean(sflip * sm_tp)) / fraction_s1,
                (np.mean(sflip * sm_fn) - np.mean(sconst * sm_fn)) / fraction_s1,
                (np.mean(oflip * om_tp) - np.mean(oconst * om_tp)) / fraction_s0,
                (np.mean(oconst * om_fn) - np.mean(oflip * om_fn)) / fraction_s0,
            ],
            [
                (np.mean(sconst * sm_fp) - np.mean(sflip * sm_fp)) / (1 - fraction_s1),
                (np.mean(sflip * sm_tn) - np.mean(sconst * sm_tn)) / (1 - fraction_s1),
                (np.mean(oflip * om_fp) - np.mean(oconst * om_fp)) / (1 - fraction_s0),
                (np.mean(oconst * om_tn) - np.mean(oflip * om_tn)) / (1 - fraction_s0),
            ],
        ]

        b_eq = [
            (np.mean(oflip * om_tp) + np.mean(oconst * om_fn)) / fraction_s0
            - (np.mean(sflip * sm_tp) + np.mean(sconst * sm_fn)) / fraction_s1,
            (np.mean(oflip * om_fp) + np.mean(oconst * om_tn)) / (1 - fraction_s0)
            - (np.mean(sflip * sm_fp) + np.mean(sconst * sm_tn)) / (1 - fraction_s1),
        ]

        # Linear program
        return linprog(coeffs, A_ub=inequalilty_constraint_matrix, b_ub=b_ub, A_eq=a_eq, b_eq=b_eq)

    def _predict(
        self, model_params: OptimizeResult, test_predictions: Prediction, test: TestTuple, seed: int
    ) -> Prediction:
        random = RandomState(seed=seed)
        sp2p, sn2p, op2p, on2p = model_params.x

        # Create boolean conditioning vectors for protected groups
        mask_s1 = test.s.to_numpy() == 1
        mask_s0 = test.s.to_numpy() == 0

        test_preds_numpy: np.ndarray = test_predictions.hard.to_numpy()

        # Randomly flip labels according to the probabilities in model_params
        self_fair_pred = test_preds_numpy[mask_s1].copy()
        self_pp_indices = (test_preds_numpy[mask_s1] == self.favorable_label).nonzero()[0]
        self_pn_indices = (test_preds_numpy[mask_s1] == self.unfavorable_label).nonzero()[0]
        random.shuffle(self_pp_indices)
        random.shuffle(self_pn_indices)

        n2p_indices = self_pn_indices[: int(len(self_pn_indices) * sn2p)]
        self_fair_pred[n2p_indices] = self.favorable_label
        p2n_indices = self_pp_indices[: int(len(self_pp_indices) * (1 - sp2p))]
        self_fair_pred[p2n_indices] = self.unfavorable_label

        othr_fair_pred = test_preds_numpy[mask_s0].copy()
        othr_pp_indices = (test_preds_numpy[mask_s0] == self.favorable_label).nonzero()[0]
        othr_pn_indices = (test_preds_numpy[mask_s0] == self.unfavorable_label).nonzero()[0]
        random.shuffle(othr_pp_indices)
        random.shuffle(othr_pn_indices)

        n2p_indices = othr_pn_indices[: int(len(othr_pn_indices) * on2p)]
        othr_fair_pred[n2p_indices] = self.favorable_label
        p2n_indices = othr_pp_indices[: int(len(othr_pp_indices) * (1 - op2p))]
        othr_fair_pred[p2n_indices] = self.unfavorable_label

        new_labels = np.zeros_like(test_preds_numpy, dtype=np.float64)
        new_labels[mask_s1] = self_fair_pred
        new_labels[mask_s0] = othr_fair_pred

        return Prediction(hard=pd.Series(new_labels))



"""For assessing the Mutual Information between s and yhat."""
from abc import ABC
from dataclasses import dataclass
from enum import Enum, auto
from typing import ClassVar

import numpy as np
from ranzen import enum_name_str, implements
from sklearn.metrics import normalized_mutual_info_score

from ethicml.utility import EvalTuple, Prediction

from .metric import Metric

__all__ = ["DependencyTarget", "NMI", "RenyiCorrelation", "Yanovich"]


@enum_name_str
class DependencyTarget(Enum):
    """The variable that is compared to the predictions in order to check how similar they are."""

    s = auto()
    y = auto()


class _DependenceMeasure(Metric, ABC):
    """Base class for dependence measures, which tell you how dependent two variables are."""

    base: DependencyTarget
    _base_name: ClassVar[str]
    apply_per_sensitive: ClassVar[bool] = False

    @implements(Metric)
    def get_name(self) -> str:
        return f"{self._base_name} preds and {self.base}"


@dataclass
class NMI(_DependenceMeasure):
    """Normalized Mutual Information.

    Also called V-Measure. Defined in this paper: https://www.aclweb.org/anthology/D07-1043.pdf
    """

    base: DependencyTarget = DependencyTarget.s
    _base_name: ClassVar[str] = "NMI"

    @implements(Metric)
    def score(self, prediction: Prediction, actual: EvalTuple) -> float:
        base_values = actual.y if self.base is DependencyTarget.y else actual.s
        return normalized_mutual_info_score(
            base_values.to_numpy().ravel(),
            prediction.hard.to_numpy().ravel(),
            average_method="arithmetic",
        )


@dataclass
class Yanovich(_DependenceMeasure):
    """Yanovich Metric. Measures how dependent two random variables are.

    As defined in this paper: https://arxiv.org/abs/1008.0492
    """

    base: DependencyTarget = DependencyTarget.s
    _base_name: ClassVar[str] = "Yanovich"

    @implements(Metric)
    def score(self, prediction: Prediction, actual: EvalTuple) -> float:
        base_values = actual.y if self.base is DependencyTarget.y else actual.s
        return self._dependence(base_values.to_numpy().ravel(), prediction.hard.to_numpy().ravel())

    @staticmethod
    def _dependence(x: np.ndarray, y: np.ndarray) -> float:
        x_vals = np.unique(x)
        y_vals = np.unique(y)
        if len(x_vals) < 2 or len(y_vals) < 2:
            return 1.0

        total = len(x)
        assert total == len(y)

        joint = np.empty((len(x_vals), len(y_vals)))

        for i, x_val in enumerate(x_vals):
            for k, y_val in enumerate(y_vals):
                # count how often x_val and y_val co-occur
                joint[i, k] = _count_true((x == x_val) & (y == y_val)) / total

        sum_sq_determinants = 0.0  # the sum of the squares of all second-order determinants
        # find all 2x2 submatrices and compute their determinant
        for i in range(len(x_vals)):
            # i < j
            for j in range(i + 1, len(x_vals)):
                for k in range(len(y_vals)):
                    # k < l
                    for l in range(k + 1, len(y_vals)):
                        determinant = joint[i, k] * joint[j, l] - joint[j, k] * joint[i, l]
                        sum_sq_determinants += determinant**2

        marginal = np.sum(joint, axis=1)

        denominator = 0.0
        for i in range(len(x_vals)):
            # i < j
            for j in range(i + 1, len(x_vals)):
                denominator += marginal[i] ** 2 * marginal[j] ** 2

        return sum_sq_determinants / denominator


@dataclass
class RenyiCorrelation(_DependenceMeasure):
    """Renyi correlation. Measures how dependent two random variables are.

    As defined in this paper: https://link.springer.com/content/pdf/10.1007/BF02024507.pdf ,
    titled "On Measures of Dependence" by Alfréd Rényi.
    """

    base: DependencyTarget = DependencyTarget.s
    _base_name: ClassVar[str] = "Renyi"

    @implements(Metric)
    def score(self, prediction: Prediction, actual: EvalTuple) -> float:
        base_values = actual.y if self.base is DependencyTarget.y else actual.s
        return self._corr(base_values.to_numpy().ravel(), prediction.hard.to_numpy().ravel())

    @staticmethod
    def _corr(x: np.ndarray, y: np.ndarray) -> float:
        x_vals = np.unique(x)
        y_vals = np.unique(y)
        if len(x_vals) < 2 or len(y_vals) < 2:
            return 1.0

        total = len(x)
        assert total == len(y)

        joint = np.empty((len(x_vals), len(y_vals)))

        for i, x_val in enumerate(x_vals):
            for k, y_val in enumerate(y_vals):
                # count how often x_val and y_val co-occur
                joint[i, k] = _count_true((x == x_val) & (y == y_val)) / total

        marginal_rows = np.sum(joint, axis=0, keepdims=True)
        marginal_cols = np.sum(joint, axis=1, keepdims=True)
        q_matrix = joint / np.sqrt(marginal_rows) / np.sqrt(marginal_cols)
        # singular value decomposition of Q
        singulars = np.linalg.svd(q_matrix, compute_uv=False)

        # return second-largest singular value
        return singulars[1]


def _count_true(mask: np.ndarray) -> int:
    """Count the number of elements that are True."""
    return np.count_nonzero(mask)



"""Class to describe features of the Adult dataset."""
from dataclasses import dataclass
from enum import Enum
from typing import ClassVar, Type, Union

from ..dataset import LoadableDataset
from ..util import LabelSpec, flatten_dict, reduce_feature_group, simple_spec

__all__ = ["Adult", "AdultSplits"]


class AdultSplits(Enum):
    """Available dataset splits for the Adult dataset."""

    SEX = "Sex"
    EDUCTAION = "Education"
    NATIONALITY = "Nationality"
    RACE = "Race"
    RACE_BINARY = "Race-Binary"
    RACE_SEX = "Race-Sex"
    CUSTOM = "Custom"


@dataclass
class Adult(LoadableDataset):
    """UCI Adult dataset.

    :param discrete_only: If True, continuous features are dropped. (Default: False)
    :param invert_s: If True, the (binary) ``s`` values are inverted. (Default: False)
    :param split: What to use as ``s``. (Default: "Sex")
    :param binarize_nationality: If True, nationality will be USA vs rest. (Default: False)
    :param binarize_race: If True, race will be white vs rest. (Default: False)
    """

    Splits: ClassVar[Type[AdultSplits]] = AdultSplits

    split: AdultSplits = Splits.SEX
    binarize_nationality: bool = False
    binarize_race: bool = False

    def __post_init__(self) -> None:
        disc_feature_groups = {
            "education": [
                "education_10th",
                "education_11th",
                "education_12th",
                "education_1st-4th",
                "education_5th-6th",
                "education_7th-8th",
                "education_9th",
                "education_Assoc-acdm",
                "education_Assoc-voc",
                "education_Bachelors",
                "education_Doctorate",
                "education_HS-grad",
                "education_Masters",
                "education_Preschool",
                "education_Prof-school",
                "education_Some-college",
            ],
            "marital-status": [
                "marital-status_Divorced",
                "marital-status_Married-AF-spouse",
                "marital-status_Married-civ-spouse",
                "marital-status_Married-spouse-absent",
                "marital-status_Never-married",
                "marital-status_Separated",
                "marital-status_Widowed",
            ],
            "native-country": [
                "native-country_Cambodia",
                "native-country_Canada",
                "native-country_China",
                "native-country_Columbia",
                "native-country_Cuba",
                "native-country_Dominican-Republic",
                "native-country_Ecuador",
                "native-country_El-Salvador",
                "native-country_England",
                "native-country_France",
                "native-country_Germany",
                "native-country_Greece",
                "native-country_Guatemala",
                "native-country_Haiti",
                "native-country_Holand-Netherlands",
                "native-country_Honduras",
                "native-country_Hong",
                "native-country_Hungary",
                "native-country_India",
                "native-country_Iran",
                "native-country_Ireland",
                "native-country_Italy",
                "native-country_Jamaica",
                "native-country_Japan",
                "native-country_Laos",
                "native-country_Mexico",
                "native-country_Nicaragua",
                "native-country_Outlying-US(Guam-USVI-etc)",
                "native-country_Peru",
                "native-country_Philippines",
                "native-country_Poland",
                "native-country_Portugal",
                "native-country_Puerto-Rico",
                "native-country_Scotland",
                "native-country_South",
                "native-country_Taiwan",
                "native-country_Thailand",
                "native-country_Trinadad&Tobago",
                "native-country_United-States",
                "native-country_Vietnam",
                "native-country_Yugoslavia",
            ],
            "occupation": [
                "occupation_Adm-clerical",
                "occupation_Armed-Forces",
                "occupation_Craft-repair",
                "occupation_Exec-managerial",
                "occupation_Farming-fishing",
                "occupation_Handlers-cleaners",
                "occupation_Machine-op-inspct",
                "occupation_Other-service",
                "occupation_Priv-house-serv",
                "occupation_Prof-specialty",
                "occupation_Protective-serv",
                "occupation_Sales",
                "occupation_Tech-support",
                "occupation_Transport-moving",
            ],
            "race": [
                "race_Amer-Indian-Eskimo",
                "race_Asian-Pac-Islander",
                "race_Black",
                "race_Other",
                "race_White",
            ],
            "relationship": [
                "relationship_Husband",
                "relationship_Not-in-family",
                "relationship_Other-relative",
                "relationship_Own-child",
                "relationship_Unmarried",
                "relationship_Wife",
            ],
            "salary": ["salary_<=50K", "salary_>50K"],
            "sex": ["sex_Female", "sex_Male"],
            "workclass": [
                "workclass_Federal-gov",
                "workclass_Local-gov",
                "workclass_Private",
                "workclass_Self-emp-inc",
                "workclass_Self-emp-not-inc",
                "workclass_State-gov",
                "workclass_Without-pay",
            ],
        }
        discrete_features = flatten_dict(disc_feature_groups)

        continuous_features = [
            "age",
            "capital-gain",
            "capital-loss",
            "education-num",
            "hours-per-week",
        ]

        sens_attr_spec: Union[str, LabelSpec]
        if self.split is AdultSplits.SEX:
            sens_attr_spec = "sex_Male"
            s_prefix = ["sex"]
            class_label_spec = "salary_>50K"
            class_label_prefix = ["salary"]
        elif self.split is AdultSplits.RACE:
            sens_attr_spec = simple_spec({"race": disc_feature_groups["race"]})
            s_prefix = ["race"]
            class_label_spec = "salary_>50K"
            class_label_prefix = ["salary"]
        elif self.split is AdultSplits.RACE_BINARY:
            sens_attr_spec = "race_White"
            s_prefix = ["race"]
            class_label_spec = "salary_>50K"
            class_label_prefix = ["salary"]
        elif self.split is AdultSplits.CUSTOM:
            sens_attr_spec = ""
            s_prefix = []
            class_label_spec = ""
            class_label_prefix = []
        elif self.split is AdultSplits.RACE_SEX:
            sens_attr_spec = simple_spec({"sex": ["sex_Male"], "race": disc_feature_groups["race"]})
            s_prefix = ["race", "sex"]
            class_label_spec = "salary_>50K"
            class_label_prefix = ["salary"]
        elif self.split is AdultSplits.NATIONALITY:
            sens = "native-country"
            sens_attr_spec = simple_spec({sens: disc_feature_groups[sens]})
            s_prefix = ["native-country"]
            class_label_spec = "salary_>50K"
            class_label_prefix = ["salary"]
        elif self.split is AdultSplits.EDUCTAION:
            to_keep = ["education_HS-grad", "education_Some-college"]
            remaining_feature_name = "other"
            discrete_features = reduce_feature_group(
                disc_feature_groups=disc_feature_groups,
                feature_group="education",
                to_keep=to_keep,
                remaining_feature_name="_" + remaining_feature_name,
            )
            sens_attr_spec = simple_spec(
                {"education": to_keep + ["education_" + remaining_feature_name]}
            )
            s_prefix = ["education"]
            class_label_spec = "salary_>50K"
            class_label_prefix = ["salary"]
        else:
            raise NotImplementedError

        name = f"Adult {self.split.value}"
        if self.binarize_nationality:
            discrete_features = reduce_feature_group(
                disc_feature_groups=disc_feature_groups,
                feature_group="native-country",
                to_keep=["native-country_United-States"],
                remaining_feature_name="_not_United-States",
            )
            if self.split is AdultSplits.SEX:
                assert len(discrete_features) == 61  # 57 (discrete) features + 4 class labels
            name += ", binary nationality"
        if self.binarize_race:
            discrete_features = reduce_feature_group(
                disc_feature_groups=disc_feature_groups,
                feature_group="race",
                to_keep=["race_White"],
                remaining_feature_name="_not_White",
            )
            if self.split is AdultSplits.SEX and self.binarize_nationality:
                assert len(discrete_features) == 58  # 54 (discrete) features + 4 class labels
            if self.split is AdultSplits.SEX and not self.binarize_nationality:
                assert len(discrete_features) == 97  # 93 (discrete) features + 4 class labels
            name += ", binary race"

        super().__init__(
            name=name,
            num_samples=45222,
            features=discrete_features + continuous_features,
            cont_features=continuous_features,
            sens_attr_spec=sens_attr_spec,
            class_label_spec=class_label_spec,
            filename_or_path="adult.csv.zip",
            s_prefix=s_prefix,
            class_label_prefix=class_label_prefix,
            discrete_only=self.discrete_only,
            discrete_feature_groups=disc_feature_groups,
        )



"""Abstract Base Class of all algorithms in the framework."""
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Tuple, TypeVar

from ethicml.algorithms.algorithm_base import Algorithm
from ethicml.utility import DataTuple, SubgroupTuple

__all__ = ["PreAlgorithm"]

T = TypeVar("T", DataTuple, SubgroupTuple)
_PA = TypeVar("_PA", bound="PreAlgorithm")


class PreAlgorithm(Algorithm, ABC):
    """Abstract Base Class for all algorithms that do pre-processing."""

    @abstractmethod
    def fit(self: _PA, train: DataTuple, seed: int = 888) -> Tuple[_PA, DataTuple]:
        """Fit transformer on the given data.

        :param train: Data tuple of the training data.
        :param seed: Random seed for model initialization.
        :returns: A tuple of Self and the test data.
        """

    @abstractmethod
    def transform(self, data: T) -> T:
        """Generate fair features with the given data.

        :param data: Data to transform.
        :returns: Transformed data.
        """

    @abstractmethod
    def run(self, train: DataTuple, test: T, seed: int = 888) -> Tuple[DataTuple, T]:
        """Generate fair features with the given data.

        :param train: Data tuple of the training data.
        :param test: Data tuple of the test data.
        :param seed: Random seed for model initialization.
        :returns: A tuple of the transforme training data and the test data.
        """

    def run_test(self, train: DataTuple, test: T, seed: int = 888) -> Tuple[DataTuple, T]:
        """Run with reduced training set so that it finishes quicker.

        :param train: Data tuple of the training data.
        :param test: Data tuple of the test data.
        :param seed: Random seed for model initialization.
        :returns: A tuple of the transforme training data and the test data.
        """
        train_testing = train.get_n_samples()
        return self.run(train_testing, test, seed)

    @abstractmethod
    def get_out_size(self) -> int:
        """Return the number of features to generate."""



"""Generate Synthetic data."""
from pathlib import Path

import numpy as np
import pandas as pd

from ethicml.data.csvs.utility import sigmoid


def main() -> None:
    r"""Make synthetic data.

    Generate synthetic data that conforms to 'Scenario 3'

    Scenario 3
    ----------

    +-----+   +-----+
    | Y_1 +<--+ X_1 +-+
    +-----+   +-----+ |   +-----+   +---+
                      +-->+ Y_3 +<--+ S |
    +-----+   +-----+ |   +-----+   +---+
    | Y_2 +<--+ X_2 +-+
    +-----+   +-----+

    In this scenario, there are two input variables, X_1 and X_2.
    There are three outcome variables, Y_1, Y_2 & Y_3.
    There is one sensitive attribute, S, which is independent of both X, Y_1 and Y_2,
    but not independent of Y_3.

    We have:
    S ~ B(0.5)

    X_1 ~ N(0, 2)
    X_2 ~ N(-1.5, 4)

    Y_1 ~ B(sigmoid(X_1)))
    Y_2 ~ B(sigmoid(X_2))
    Y_3 ~ B(sigmoid(((X_1 + X_2)/2)+S))
    """
    seed = 0
    samples = 100_000

    np.random.seed(seed)

    s = np.random.binomial(1, 0.6, samples)

    x_1f = np.random.normal(0, 0.5, samples)
    x_1 = x_1f
    x_2 = np.random.normal(-1, 3, samples)
    x_2f = x_2

    y_1 = np.random.binomial(1, sigmoid(x_1))
    y_1f = np.random.binomial(1, sigmoid(x_1f))
    y_2 = np.random.binomial(1, sigmoid(x_2))
    y_2f = y_2
    p: np.ndarray = x_1 + x_2
    pf: np.ndarray = x_1f + x_2f
    y_3 = np.random.binomial(1, sigmoid(p + 0.5 * (s * 2 - 1)))
    y_3f = np.random.binomial(1, sigmoid(pf))

    noise_1 = np.random.normal(0, 4, samples)
    noise_2 = np.random.normal(3, 7, samples)

    print(
        s.mean(),
        x_1.mean(),
        x_1f.mean(),
        x_2.mean(),
        x_2f.mean(),
        y_1.mean(),
        y_1f.mean(),
        y_2.mean(),
        y_2f.mean(),
        y_3.mean(),
        y_3f.mean(),
    )

    df = pd.DataFrame(
        data={
            "x1": x_1,
            "x1f": x_1f,
            "x2": x_2,
            "x2f": x_2f,
            "s": s,
            "y1": y_1,
            "y1f": y_1f,
            "y2": y_2,
            "y2f": y_2f,
            "y3": y_3,
            "y3f": y_3f,
            "n1": noise_1,
            "n2": noise_2,
        }
    )

    # Shuffle the data,
    df = df.sample(frac=1.0, random_state=seed).reset_index(drop=True)

    # Save the CSV
    df.to_csv(str(Path(__file__).parent / "synthetic_scenario_3.csv"), index=False)


if __name__ == "__main__":
    main()



"""Demographic Parity Label flipping approach."""
from dataclasses import dataclass
from typing import Tuple

import numpy as np
import pandas as pd
from ranzen import implements

from ethicml.utility import DataTuple, Prediction, TestTuple

from .post_algorithm import PostAlgorithm

__all__ = ["DPFlip"]


@dataclass
class DPFlip(PostAlgorithm):
    """Randomly flip a number of decisions such that perfect demographic parity is achieved."""

    @implements(PostAlgorithm)
    def get_name(self) -> str:
        return "DemPar. Post Process"

    @implements(PostAlgorithm)
    def fit(self, train_predictions: Prediction, train: DataTuple) -> "DPFlip":
        return self

    @implements(PostAlgorithm)
    def predict(self, test_predictions: Prediction, test: TestTuple, seed: int = 888) -> Prediction:
        x, y = self._fit(test, test_predictions)
        _test_preds = self._flip(
            test_predictions, test, flip_0_to_1=True, num_to_flip=x, s_group=0, seed=seed
        )
        return self._flip(_test_preds, test, flip_0_to_1=False, num_to_flip=y, s_group=1, seed=seed)

    @implements(PostAlgorithm)
    def run(
        self,
        train_predictions: Prediction,
        train: DataTuple,
        test_predictions: Prediction,
        test: TestTuple,
        seed: int = 888,
    ) -> Prediction:
        x, y = self._fit(test, test_predictions)
        _test_preds = self._flip(
            test_predictions, test, flip_0_to_1=True, num_to_flip=x, s_group=0, seed=seed
        )
        return self._flip(_test_preds, test, flip_0_to_1=False, num_to_flip=y, s_group=1, seed=seed)

    @staticmethod
    def _flip(
        preds: Prediction,
        dt: TestTuple,
        flip_0_to_1: bool,
        num_to_flip: int,
        s_group: int,
        seed: int,
    ) -> Prediction:
        if num_to_flip >= 0:
            pre_y_val = 0 if flip_0_to_1 else 1
            post_y_val = 1 if flip_0_to_1 else 0
        else:
            pre_y_val = 1 if flip_0_to_1 else 0
            post_y_val = 0 if flip_0_to_1 else 1
            num_to_flip = abs(num_to_flip)

        _y = preds.hard[preds.hard == pre_y_val]
        _s = preds.hard[dt.s == s_group]
        idx_s_y = _y.index.intersection(_s.index)
        rng = np.random.RandomState(seed)
        idxs = list(rng.permutation(idx_s_y))
        update = pd.Series({idx: post_y_val for idx in idxs[:num_to_flip]}, dtype=preds.hard.dtype)
        preds.hard.update(update)
        return preds

    @staticmethod
    def _fit(test: TestTuple, preds: Prediction) -> Tuple[int, int]:
        y_0 = preds.hard[preds.hard == 0]
        y_1 = preds.hard[preds.hard == 1]
        s_0 = test.s[test.s == 0]
        s_1 = test.s[test.s == 1]
        # Naming is nSY
        n00 = preds.hard[s_0.index.intersection(y_0.index)].count()
        n01 = preds.hard[s_0.index.intersection(y_1.index)].count()
        n10 = preds.hard[s_1.index.intersection(y_0.index)].count()
        n11 = preds.hard[s_1.index.intersection(y_1.index)].count()

        a = (((n00 + n01) * n11) - ((n10 + n11) * n01)) / (n00 + n01)
        b = (n10 + n11) / (n00 + n01)

        if b > 1:
            x = a / b
            z = 0.0
        else:
            x = 0.0
            z = a
        return int(round(x)), int(round(z))
